package decac

import jllvm.LLVMType
import jllvm.LLVMStructType
import jllvm.LLVMIntegerType

//Case analysis of the different kinds of types passed as type annotations for members.
abstract class MemberDeclarationType {
  def memberType(owner: Scope): MemberType
}
case class SimpleDeclarationType(str: String) {
  def memberType(owner: Scope): MemberType = owner.lookup(str) match {
    case actual: ActualType => new Actual(actual)
    case arg: TypeArgument => new Argument(arg)
    case classType: ClassType => new ClassConstructionMember(classType,Nil)
    case _ => throw new Exception("Given name for type of class member does not denote a type or type argument.")
  }
}
case class ComplexDeclarationType(constructor: String,args: List[MemberDeclarationType]) {
  def memberType(owner: Scope): MemberType = owner.lookup(constructor) match {
    case classType: ClassType => new ClassConstructionMember(classType,args.map(arg => arg.memberType(owner)))
    case _ => throw new Exception("Given name for a parameterized class in member declaration does not denote a class.")
  }
}

class ClassMemberDeclaration(n: Option[String],num: Option[Int],memType: MemberDeclarationType) {
  val memberType = memType
  val name: Option[String] = n
  val number: Int = num match { case Some(i) => i case None => 1 }
}

class ClassScope(c: ClassType,p: Scope) extends Scope(p) {
  val ownerClass = c
  override val parent: Scope = p match { case c: ClassScope => c case m: Module => m case _ => throw new Exception("Passed a Scope that is neither a ClassScope nor a Module to constructing a ClassScope.") }
  val module: Option[Module] = p match { case m: Module => Some(m) case c: ClassScope => None case _ => throw new Exception("Passed a Scope that is neither a ClassScope nor a Module to constructing a ClassScope.") }
}

abstract class MemberType
case class Actual(t: ActualType) extends MemberType
case class Argument(a: TypeArgument) extends MemberType
case class ClassConstructionMember(c: ClassType,args: List[MemberType]) extends MemberType

class ClassMember(c: ClassType,decl: ClassMemberDeclaration) extends Scopeable {
  //TODO: Need a better way to generate names for anonymous members.
  val name: String = decl.name match { case Some(str) => str case None => number.toString }
  val scope: ClassScope = { c.scope.declare(this); c.scope }
  val number: Int = decl.number

  val memberType: MemberType = decl.memberType.memberType(scope)
}

class ClassType(n: String,p: Option[MemberDeclarationType],m: Option[Module],args: List[String],mems: List[ClassMemberDeclaration]) extends Scopeable {
  val name: String = n
  val owner: Module = m match { case Some(mod) => {mod.declare(this); mod} case None => throw new Exception("Every class must be declared in a module.") }
  val parent: Option[MemberType] = p match {
    case None => None
    case Some(p) => Some(p.memberType(this.scope))
  }
  val scope: ClassScope = new ClassScope(this,owner)
  val arguments: List[TypeArgument] = args.map(arg => new TypeArgument(this,arg))
  val members: List[ClassMember] = mems.map(mem => new ClassMember(this,mem))
  
  def subtypes(c: ClassType): Boolean = c == this || (parent match {
    case None => false
    case Some(parentmt) => parentmt match {
      case ClassConstructionMember(parclass,args) => parclass == c || parclass.subtypes(c)
      case _ => false
    }
  })
  
  def identicalTypeArguments(c: ClassType): List[Tuple2[Int,Int]] = {
    arguments.zipWithIndex.zip(c.arguments.zipWithIndex).filter(pair => pair._1._1 == pair._2._1).map(pair => (pair._1._2,pair._2._2))
  }
  
  def commonSuperclass(c: ClassType): Option[MemberType] = {
    if(subtypes(c))
      Some(new ClassConstructionMember(c,c.arguments.map(arg => Argument(arg))))
    else if(c.subtypes(this))
      Some(new ClassConstructionMember(this,arguments.map(arg => Argument(arg))))
    else parent match {
      case None => None
      case Some(parent) => parent match {
        case ClassConstructionMember(parent,args) => parent.commonSuperclass(c)
        case Argument(arg) => c.parent match {
          case Some(Argument(a)) => Some(parent)
          case _ => None
        }
        case Actual(actual) => throw new Exception("Class types cannot subtype non-class types.")
      }
    }
  }
}

class TypeArgument(ct: ClassType,n: String) extends TypeExpression with Scopeable {
  val owner: ClassType = ct
  val name: String = n
  val scope: ClassScope = { ct.scope.declare(this) ; ct.scope }
  
  override def subtypes(e: TypeExpression): Boolean = false
  override def equals(e: TypeExpression): Boolean = false
}

class ClassConstruction(c: ClassType) extends TypeConstruction {
  val constructionClass: ClassType = c
  override var contravariants: List[TypeExpression] = constructionClass.arguments.map(arg => new TypeVariableExpression(new TypeVariable,Handedness.LeftValue))
  override var covariants: List[TypeExpression] = Nil
  override def constructorSubtypes(c: TypeConstruction): Boolean = c match {
    case cc: ClassConstruction => constructionClass.subtypes(cc.constructionClass)
    case _ => false
  }
  override def constructorEquals(c: TypeConstruction): Boolean = c match {
    case cc: ClassConstruction => cc.constructionClass == constructionClass
    case _ => false
  }
  override def superCovariants(c: TypeConstruction): List[Tuple2[TypeExpression,TypeExpression]] = Nil
  override def superContravariants(c: TypeConstruction): List[Tuple2[TypeExpression,TypeExpression]] = c match {
    case cc: ClassConstruction => if(constructionClass.subtypes(cc.constructionClass)) constructionClass.identicalTypeArguments(cc.constructionClass).map(pair => (contravariants.apply(pair._1),cc.contravariants.apply(pair._2))) else Nil
    case _ => Nil
  }
  
  protected def typeArgument(arg: TypeArgument): TypeExpression = constructionClass.arguments.zipWithIndex.find(pair => arg == pair._1) match {
    case None => throw new Exception("Unspecified type argument " + arg.name + " in specialization of generic class " + constructionClass.name + ".")
    case Some(pair) => contravariants.apply(pair._2)
  }
  
  protected def typeMemberType(memType: MemberType): TypeExpression = memType match {
    case Actual(t) => t
    case Argument(a) => typeArgument(a)
    case ClassConstructionMember(ct,args) => {
      val result = new ClassConstruction(ct)
      result.contravariants = args.map(arg => typeMemberType(arg))
      result
    }
  }
 
  def typeMember(member: ClassMember): TypeExpression = {
    if(constructionClass.scope.lookup(member.name) != member)
      throw new Exception("Given member was not found in the class scope that was supposed to type it.")
    else
      typeMemberType(member.memberType)
  }
  
  override def makeCommonSupertype(c: TypeConstruction,constraints: Constraints): Option[TypeExpression] = c match {
    case cc: ClassConstruction => {
      val commonSuper: Option[MemberType] = constructionClass.commonSuperclass(cc.constructionClass)
      commonSuper match {
        case None => None
        case Some(superClass) => superClass match {
          case ClassConstructionMember(superClass,args) => {
            val result = new ClassConstruction(superClass)
            constraints.push(new LesserEq(this,result))
            constraints.push(new LesserEq(c,result))
            Some(result)
          }
          case Argument(arg) => {
            val result = typeArgument(arg)
            val parentType = cc.constructionClass.parent match { case Some(Argument(a)) => cc.typeArgument(arg) case _ => return None }
            constraints.push(new LesserEq(this,result))
            constraints.push(new Equal(parentType,result))
            constraints.push(new LesserEq(c,parentType))
            Some(result)
          }
        }
      }
    }
    case _ => None
  }
      
  private def expressMember(substitution: Substitution,memberType: MemberType): TypeConstruction = memberType match {
    case Actual(actual) => actual
    case Argument(arg) => substitution.solve(typeArgument(arg))
    case ClassConstructionMember(c,args) => {
      val cc = new ClassConstruction(c)
      cc.contravariants = args.map(arg => expressMember(substitution,arg))
      return cc
    }
  }
  
  private def compileMember(substitution: Substitution,memberType: MemberType): LLVMType = {
    expressMember(substitution,memberType).compile(substitution)
  }
  
  override def compile(substitution: Substitution): LLVMStructType = {
    val memberTypes: List[LLVMType] = constructionClass.members.map(member => compileMember(substitution,member.memberType))
    constructionClass.parent match {
      //TODO: Add a way for programmers to specify whether they want packed or unpacked classes.
      case None => new LLVMStructType(((new LLVMIntegerType(32)) :: memberTypes).toArray,true)
      case Some(parent) => new LLVMStructType(((new LLVMIntegerType(32)) :: compileMember(substitution,parent) :: memberTypes).toArray,true)
    }
  }
}
