package decac;

import jllvm.LLVMType;

abstract class TypeExpression {
  def subtypes(e: TypeExpression): Boolean
  def equals(e: TypeExpression): Boolean = e == this
}

class TypeVariable {
  def subtypes(e: TypeExpression): Boolean = {
    equals(e)
  }
  
  def equals(e: TypeExpression): Boolean = e match {
    case evar: TypeVariableExpression => evar.variable == this
    case _ => false
  }
  
  def tightenUpperBound(newbound: TypeConstruction,constraints: Constraints,substitution: Substitution): TypeRange = new TypeRange(None,Some(newbound))
  def tightenLowerBound(newbound: TypeConstruction,constraints: Constraints,substitution: Substitution): TypeRange = new TypeRange(Some(newbound),None)
}

class TypeRange(lower: Option[TypeConstruction],upper: Option[TypeConstruction]) extends TypeVariable {
  protected var lower_bound: Option[TypeConstruction] = { assert(lower != None || upper != None) ; lower }
  protected var upper_bound: Option[TypeConstruction] = { assert(lower != None || upper != None) ; upper }
  
  def upperBound: TypeConstruction = upper_bound match {
    case Some(upper) => upper
    case None => lower_bound match {
      case Some(lower) => lower
      case None => throw new Exception("A type range must have an upper or lower bound.")
    }
  }
  
  def lowerBound: TypeConstruction = lower_bound match {
    case Some(lower) => lower
    case None => upper_bound match {
      case Some(upper) => upper
      case None => throw new Exception("A type range must have an upper or lower bound.")
    }
  }
  
  override def tightenUpperBound(newbound: TypeConstruction,constraints: Constraints,substitution: Substitution): TypeRange = {
    val newUpper = upper_bound match {
      case Some(upper) => {
        if(newbound.constructorEquals(upper)) {
          upper.covariants.zip(newbound.covariants).map(pair => (new Equal(pair._1,pair._2)).evaluate(constraints,substitution))
          upper.contravariants.zip(newbound.contravariants).map(pair => (new Equal(pair._1,pair._2)).evaluate(constraints,substitution))
          upper
        }
        else if(newbound.constructorSubtypes(upper))
          newbound
        else if(upper.constructorSubtypes(newbound))
          upper
        else
          throw new Exception("Cannot tighten the upper bound " + upper + " of type variable with unrelated type " + newbound + ".")
        }
      case None => newbound
    }
    new TypeRange(lower_bound,Some(newUpper))
  }
  
  override def tightenLowerBound(newbound: TypeConstruction,constraints: Constraints,substitution: Substitution): TypeRange = {
    val newLower: TypeConstruction = lower_bound match {
      case Some(lower) => {
        if(newbound.constructorEquals(lower)) {
          lower.covariants.zip(newbound.covariants).map(pair => (new Equal(pair._1,pair._2)).evaluate(constraints,substitution))
          lower.contravariants.zip(newbound.contravariants).map(pair => (new Equal(pair._1,pair._2)).evaluate(constraints,substitution))
          lower
        }
        else lower.makeCommonSupertype(newbound,constraints) match {
          case Some(bound) => {
            (new LesserEq(lower,bound)).evaluate(constraints,substitution)
            (new LesserEq(newbound,bound)).evaluate(constraints,substitution)
            substitution.solve(bound)
          }
          case None => throw new Exception("Cannot tighten the lower bound " + lower + " of type variable with unrelated type " + newbound + ".")
        }
      }
      case None => newbound
    }
    new TypeRange(Some(newLower),upper_bound)
  }
  
  override def subtypes(e: TypeExpression): Boolean = e match {
    case evar: TypeVariableExpression => evar.variable match {
      case erange: TypeRange => upperBound.subtypes(erange.lowerBound)
      case evar: TypeVariable => false
    }
    case econst: TypeConstruction => upperBound.subtypes(econst)
  }
  
  override def equals(e: TypeExpression): Boolean = e match {
    case evar: TypeVariableExpression => evar.variable match {
      case erange: TypeRange => upperBound.equals(erange.upperBound) && lowerBound.equals(erange.lowerBound)
      case evar: TypeVariable => false
    }
    case econst: TypeConstruction => upperBound.equals(econst) && lowerBound.equals(econst)
  }
}

object Handedness extends Enumeration {
  type Handedness = Value
  val RightValue = Value("RightValue")
  val LeftValue = Value("LeftValue")
}
import Handedness._

class TypeVariableExpression(tvar: TypeVariable,hand: Handedness) extends TypeExpression {
  val handedness: Handedness = hand
  val variable: TypeVariable = tvar
  
  override def subtypes(e: TypeExpression): Boolean = { 
    variable.subtypes(e) 
  }
  
  override def equals(e: TypeExpression): Boolean = {
    variable.equals(e)
  }
  
  def tightenUpperBound(newbound: TypeConstruction,constraints: Constraints,substitution: Substitution): TypeVariableExpression = {
    new TypeVariableExpression(variable.tightenUpperBound(newbound,constraints,substitution),handedness)
  }
  
  def tightenLowerBound(newbound: TypeConstruction,constraints: Constraints,substitution: Substitution): TypeVariableExpression = {
    new TypeVariableExpression(variable.tightenLowerBound(newbound,constraints,substitution),handedness)
  }
  
  def value: Option[TypeConstruction] = variable match {
    case range: TypeRange => handedness match {
      case RightValue => Some(range.lowerBound)
      case LeftValue => Some(range.upperBound)
    }
    case tvar: TypeVariable => None
  }  
}

abstract class TypeConstruction extends TypeExpression {
  var contravariants: List[TypeExpression]
  var covariants: List[TypeExpression]
  protected def children: List[TypeExpression] = {
    contravariants ::: covariants
  }
  
  def contains(v: TypeVariableExpression): Boolean = children.exists(free => free match { case freevar: TypeVariableExpression => freevar.variable == v.variable case _ => false })
  def substituteChildren(v: TypeVariableExpression,e: TypeExpression) = v.handedness match {
    case Handedness.LeftValue => contravariants = contravariants.map(child => if(v == child) e else child)
    case Handedness.RightValue => covariants = covariants.map(child => if(v == child) e else child)
  }
  
  def constructorEquals(c: TypeConstruction): Boolean
  def constructorSubtypes(c: TypeConstruction): Boolean
  def superContravariants(c: TypeConstruction): List[Tuple2[TypeExpression,TypeExpression]]
  def superCovariants(c: TypeConstruction): List[Tuple2[TypeExpression,TypeExpression]]
  
  def commonSupertype(c: TypeConstruction): Option[TypeConstruction] = {
    if(c.subtypes(this))
      Some(this)
    else if(subtypes(c))
      Some(c)
    else
      None
  }
  
  def makeCommonSupertype(c: TypeConstruction,constraints: Constraints): Option[TypeExpression] = {
    if(c.constructorSubtypes(this))
      Some(this)
    else if(constructorSubtypes(c))
      Some(c)
    else
      None
  }
  
  override def subtypes(e: TypeExpression): Boolean = e match {
    case actual: ActualType => constructorSubtypes(actual)
    case constr: TypeConstruction => {
      val and = (x: Boolean,y: Boolean) => x && y
      (constructorSubtypes(constr) && superContravariants(constr).map(pair => pair._2.subtypes(pair._1)).foldRight(true)(and) && superCovariants(constr).map(pair => pair._1.subtypes(pair._2)).foldRight(true)(and)) ||
      (constructorEquals(constr) && contravariants.zip(constr.contravariants).map(pair => pair._2.subtypes(pair._1)).foldRight(true)(and) && covariants.zip(constr.covariants).map(pair => pair._1.subtypes(pair._2)).foldRight(true)(and))
    }
    case evar: TypeVariableExpression => evar.variable match {
      case erange: TypeRange => subtypes(erange.lowerBound)
      case evar: TypeVariable => false
    }
    case _ => false
  }
  
  override def equals(e: TypeExpression): Boolean = e match {
    case constr: TypeConstruction => {
      val and = (x: Boolean,y: Boolean) => x && y
      constructorEquals(constr) &&
      contravariants.zip(constr.contravariants).map(pair => pair._2.equals(pair._1)).foldRight(true)(and) &&
      covariants.zip(constr.covariants).map(pair => pair._1.equals(pair._2)).foldRight(true)(and)
    }
    case evar: TypeVariableExpression => evar.variable match {
      case erange: TypeRange => equals(erange.lowerBound) && equals(erange.upperBound)
      case evar: TypeVariable => false
    }
    case _ => false
  } 
  
  def compile(substitution: Substitution): LLVMType
}

abstract class ActualType(n: String,m: Option[Module]) extends TypeConstruction with Scopeable {
  val name: String = n
  val owner: Module = m match { case Some(mod) => {mod.declare(this); mod} case None => GlobalScope }
  override def scope: Module = owner	
  
  override var contravariants: List[TypeExpression] = Nil
  override var covariants: List[TypeExpression] = Nil
  override def constructorEquals(c: TypeConstruction): Boolean = c == this
  override def constructorSubtypes(c: TypeConstruction): Boolean = subtypes(c)
  override def superContravariants(c: TypeConstruction): List[Tuple2[TypeExpression,TypeExpression]] = Nil
  override def superCovariants(c: TypeConstruction): List[Tuple2[TypeExpression,TypeExpression]] = Nil
  override def equals(e: TypeExpression): Boolean = e == this
  
  def compile(substitution: Substitution): LLVMType
}
