package decac;

import scala.collection.mutable.Queue
import scala.collection.mutable.Stack

class SubstitutionElement(x: TypeExpression,y: TypeExpression) {
  val initial = x
  val substituted = y
}

class Substitution {
  val queue: Queue[SubstitutionElement] = new Queue[SubstitutionElement]()
  
  def substitute(x: TypeVariableExpression,y: TypeExpression) = y match {
    case y: TypeVariableExpression => queue.enqueue(new SubstitutionElement(x,new TypeVariableExpression(y.variable,x.handedness)))
    case _ => queue.enqueue(new SubstitutionElement(x,y))
  }
  
  def solve(x: TypeExpression): TypeConstruction = {
    var solution: TypeExpression = x
    queue.foreach(elmnt => if(elmnt.initial.equals(solution)) solution = elmnt.substituted)
    solution match {
      case solution: TypeConstruction => solution
      case _ => throw new Exception("Inferred type is neither a parameterized type nor a simple type.")
    }
  }
}

class Constraints {
  val stack: Stack[Constraint] = new Stack[Constraint]()
  
  def substitute(x: TypeVariableExpression,y: TypeExpression) = y match {
    case y: TypeVariableExpression => stack.map(constraint => constraint.substitute(x,new TypeVariableExpression(y.variable,x.handedness)))
    case _ => stack.map(constraint => constraint.substitute(x,y))
  }
  
  def push(c: Constraint): Unit = stack.push(c)
  
  def pop: Constraint = stack.pop
  
  def isEmpty = stack.isEmpty
}

class Unification {
  val constraints = new Constraints()
  
  def constrain(c: Constraint) = {
    constraints.push(c)
  }
  
  def solve: Substitution = {
    val result = new Substitution
    while(constraints.isEmpty != true)
      constraints.pop.evaluate(constraints,result)
    return result
  }
}

abstract class Constraint(x: TypeExpression,y: TypeExpression) {
  var alpha = x
  var beta = y
  
  def substitute(from: TypeVariableExpression,to: TypeExpression) = {
    alpha match {
      case a: TypeVariableExpression => if(a.equals(from)) alpha = to
      case a: TypeConstruction => a.substituteChildren(from,to)
    }
    beta match {
      case b: TypeVariableExpression => if(b.equals(from)) beta = to
      case b: TypeConstruction => b.substituteChildren(from,to)
    }
  }
  
  def evaluate(constraints: Constraints,substitution: Substitution)

  protected def map[T](f: (T,T) => Unit,x: List[T],y: List[T]): Unit = {
    if(x != Nil && y != Nil) {
      f(x.head,y.head);
      map(f,x.tail,y.tail)
    }
  }
}

case class LesserEq(x: TypeExpression,y: TypeExpression) extends Constraint(x,y) {
  //Gorramit I want my multiple-dispatch!  Hand-writing these vast case analyses sucks!
  override def evaluate(constraints: Constraints,substitution: Substitution) = alpha match {
    //TODO: What do I do when alpha or beta is a MemberTypeExpression whose variable is unknown?
    case alpha: TypeVariableExpression => beta match {
      case beta: TypeVariableExpression => {
        if(!alpha.equals(beta)) alpha.variable match {
          case alpharange: TypeRange => beta.variable match {
            case betarange: TypeRange => {
              val alphaPrime = alpha.tightenUpperBound(betarange.lowerBound,constraints,substitution)
              substitution.substitute(alpha,alphaPrime)
              constraints.substitute(alpha,alphaPrime)
              val betaPrime = beta.tightenLowerBound(alpharange.upperBound,constraints,substitution)
              substitution.substitute(beta,betaPrime)
              constraints.substitute(beta,betaPrime)
            }
            case betavar: TypeVariable => {
              val betaPrime = beta.tightenLowerBound(alpharange.upperBound,constraints,substitution)
              substitution.substitute(beta,betaPrime)
              constraints.substitute(beta,betaPrime)
            }
          }
          case alphavar: TypeVariable => beta.variable match {
            case betarange: TypeRange => {
              val alphaPrime = alpha.tightenUpperBound(betarange.lowerBound,constraints,substitution)
              substitution.substitute(alpha,alphaPrime)
              constraints.substitute(alpha,alphaPrime)
            }
            case betavar: TypeVariable => (new Equal(alpha,beta)).evaluate(constraints,substitution)
          }
        }
      }
      case beta: TypeConstruction => {
        assert(!beta.contains(alpha))
        val alphaPrime = alpha.tightenUpperBound(beta,constraints,substitution)
        substitution.substitute(alpha,alphaPrime)
        constraints.substitute(alpha,alphaPrime)
      }
    }
    case alpha: TypeConstruction => beta match {
      case beta: TypeVariableExpression => {
        assert(!alpha.contains(beta),"Type variable occurs as an infinite type recursion.")
        val betaPrime = beta.tightenLowerBound(alpha,constraints,substitution)
        substitution.substitute(beta,betaPrime)
        constraints.substitute(beta,betaPrime)
      }
      case beta: TypeConstruction => {
        if(alpha.subtypes(beta)) {
          //TODO: This works for subtyping constructions in which the child types all need to be covariant.
          //Functions need their argument types to be contravariant and the results to be covariant.
          map((alphakid: TypeExpression,betakid: TypeExpression) => constraints.push(new LesserEq(betakid,alphakid)),alpha.contravariants,beta.contravariants)
          map((alphakid: TypeExpression,betakid: TypeExpression) => constraints.push(new LesserEq(alphakid,betakid)),alpha.covariants,beta.covariants)
        }
        else
          throw new Exception("Type " + alpha + " is not a subtype of or equal to type " + beta + ".  Type inference fails.")
      }
    }
  }
}

case class Equal(x: TypeExpression,y: TypeExpression) extends Constraint(x,y) {
  //Once again, I really, really wish I could define this using multimethods.
  override def evaluate(constraints: Constraints,substitution: Substitution) = alpha match {
    case alpha: TypeVariableExpression => {
      assert(alpha != beta && (beta match { case beta: TypeConstruction => !beta.contains(alpha) case _ => true }))
      beta match {
        case beta: TypeVariableExpression => alpha.variable match {
          case arange: TypeRange => beta.variable match {
            case brange: TypeRange => {
              (new Equal(arange.upperBound,brange.upperBound)).evaluate(constraints,substitution)
              (new Equal(arange.lowerBound,brange.lowerBound)).evaluate(constraints,substitution)
            }
            case bvar: TypeVariable => {
              constraints.substitute(beta,alpha)
              substitution.substitute(beta,alpha)
            }
          }
          case avar: TypeVariable => {
            constraints.substitute(alpha,beta)
            substitution.substitute(alpha,beta)
          }
        }
      }
      substitution.substitute(alpha,beta)
      constraints.substitute(alpha,beta)
    }
    case alpha: TypeConstruction => beta match {
      case beta: TypeConstruction => {
        assert(alpha.constructorEquals(beta),"Cannot unify type expressions with different type constructors.")
        val alphacovs = alpha.covariants
        val betacovs = beta.covariants
        assert(alphacovs.length == betacovs.length,"Cannot unify type expressions with different numbers of sub-expressions.")
        map((alphakid: TypeExpression,betakid: TypeExpression) => constraints.push(new Equal(alphakid,betakid)),alphacovs,betacovs)
        val alphacons = alpha.contravariants
        val betacons = beta.contravariants
        assert(alphacons.length == betacons.length,"Cannot unify type expressions with different numbers of sub-expressions.")
        map((alphakid: TypeExpression,betakid: TypeExpression) => constraints.push(new Equal(alphakid,betakid)),alphacons,betacons)
      }
      case beta: TypeVariableExpression => {
        assert(!alpha.contains(beta))
        assert(beta.variable match {case range: TypeRange => range.lowerBound.subtypes(alpha) && alpha.subtypes(range.upperBound) case tvar: TypeVariable => true },"Cannot unify constructed type expression to the given constrained variable.")
        substitution.substitute(beta,alpha)
        constraints.substitute(beta,alpha)
      }
      case _ => throw new Exception("Cannot unify complex type expression with simple one.")
    }
  }
}
